# -*- coding: utf-8 -*-
"""
Created on Thu Jan 12 11:58:09 2017

@author: Mic
"""
from __future__ import division
import numpy as np
from numpy.linalg import norm
from wiselib2.must import *
import inspect

from matplotlib import pyplot as plt

from _ast import arg
#================================================================
#  ErrMsg
#================================================================
class ErrMsg:
    @staticmethod
    def NoPropertySetAllowed(More = ''):
        raise ValueError('Property set metod is not allowed.\n%s' % More)
        
    def InvalidInputSet(More=''):
        raise ValueError('Ivalid input parameter set used.' % More)


#================================================================
#  CLASS Debug
#================================================================        
class Debug():
    import inspect
    On = True
    _OldValue = True

        
    def Print(Str = '', NIndent = 0, Header = False):
        if Debug.On : 
            Str = (NIndent * '\t' + '%s') % Str
            Str= Str.replace('\n','\n' + NIndent * '\t')
            Str = Str if Header == False else (30 * '=-' + '\n' + Str + '\n' + 30 * '=-'+'\n') 
            print(Str)        
        
    def print(Str ='', NIndent = 0, Header = False):
        if Debug.On :
            Debug.Print(Str, NIndent, Header)
            
    def pr(LocalVarName, More = ''):
        frame = inspect.currentframe()
        if Debug.On :
            try:
                print(LocalVarName + ':= ' + str(frame.f_back.f_locals[LocalVarName]) + ' ' + More)
            except:
                pass
            finally:
                del frame
                
    def pv(LocalVarName, NIndent = 0, More = ''):
        if Debug.On :
            frame = inspect.currentframe()
            Tab = NIndent * '\t'
            if True:
                try:
                    print(Tab + LocalVarName + ':= ' + str(frame.f_back.f_locals[LocalVarName]) + ' ' + More)
                except:
                    pass
                finally:
                    del frame
                    
    

        
#================================================================
#  IsArray
#================================================================
def IsArray(x):
    return isinstance(x,np.ndarray)

#================================================================
#  IsScalar
#================================================================
def IsScalar(x):
    return isinstance(x,float) or isinstance(x,int)    
            
        
#================================================================
#  CheckArg
#================================================================
def CheckArg(NotNoneArgs, NoneArgs = []):
    '''
    Returns if ALL the values in ArgList are not none.

    I use it for initializing object which admit different paramenters, e.g.

    a = Point(x,y)    # in rectangular coordinates
    b = Point(r,phi)   # in polar cooordinates.

    '''

    a = np.all([arg is not None for arg in NotNoneArgs])

    if len(NoneArgs) == 0:
        return a
    else:
        return a and np.all([arg is None for arg in NoneArgs])
        #return a and (NoneArgs is None)

def Oversample(x = np.array([]), N=1):
    '''
    Adds sampling points in the x vector.
    E.g. Oversample([0,1], N=1) => [0, 0.5,1]
    Preserves x[0] and x[end]
    '''
    x = np.array([])
    for i, xi in enumerate(x):
        x = np.append()
#================================
#  SamplingOverLine
#================================    
def SamplingAlongLine(Theta, XYCentre, Length, NSamples = None, Step = None):
    '''
    Samples a line of angle Theta centret at XYCentre with a given Step.

    I created the function for properly sampling a plane or elliptical mirror.
    It makes sense to use this one for sampling an elliptical mirror if the 
    figure error is measured with an LTP profilometer (uniform spacing along
    a railw)
    Either the total Length or the Number of samples NSamples must be specified
        
    Parameters
    ----------
    Theta : 
        angle (radians)
    XYCentre :  2d array
        (x,y) centre of the line
    Step : float
        distance between two samples (along the linge)
    Length : float
        total length to cover (along the line)    
    NSamples : int
        (alternative to Length), number of samples

    '''
    
    m = np.tan(Theta)    
    q = XYCentre[1] - m*XYCentre[0]
    L_ = abs(Length * np.cos(Theta)    )
    x_start = XYCentre[0] - L_/2
    x_end = XYCentre[0] + L_/2
    
    if  Step!= None and Step >0:        
        N = Length/Step
    elif NSamples !=None and NSamples > 0:  
        N = NSamples
        
    x = np.linspace(x_start, x_end, N)    
    y = m*x+q
    return x,y
#================================================================
#  RMat
#================================================================
def RMat(Theta):
    '''
    Returns the rotation matrix for an angle Theta.
    '''
    return [[np.cos(Theta), -np.sin(Theta)],
                [np.sin(Theta), np.cos(Theta)]]

#================================================================
#  RotXY
#================================================================
def RotXY(x,y, Theta = 0, CentreOfRotation = np.array([0,0])):
    '''
    Rotates the arrays x (1d) and y (1d) of Theta AROUND the CentreOfRotation

    Parameters
    ----------------
    x : 1d array
        x coordinates
    y : 1d array
        y coordinates
    Theta : scalar (rad)
        Rotation angle
    CentreOfRotation : [x,y]
        Point around which the rotation is performed.By default is set to [0,0]

    Returns
    -----------------
    x : rotated x

    y : rotated y

    Examples
    ----------------
    '''
    if Theta == 0:
        return (np.array(x), np.array(y))
    Theta = -Theta # non so perché il -1, odio le matrici di rotazione.
    Vxy = np.column_stack((x,y))
    U  = np.dot(Vxy - CentreOfRotation, RMat(Theta)) + CentreOfRotation
    return (U[:,0], U[:,1])

#================================================================
#  RotPoly
#================================================================
def RotPoly(P, NewOrigin = np.array([0,0]), Angle = 0, Deg = False):
    '''
    Returns the coefficients of the rotated polynomial in the
    Parameters
    ----------------------
    P : 1d-array
        Polynomial coefficients [x^n .... n^0]
    NewOrigin : 1x2 array
        [x,y] of the new origin
    Angle : scalr ( rad)
        Rotation Angle
    '''
    P = np.array(P)
    N =   len(P)-1 # degree of the polynomial
    if N <1:
        print('Errror: Polynomial order too low (<1), finding coefficient is useles...')
        return None

    x = np.linspace(0,N, N+1)
    y = np.polyval(P,x)

    x_new, y_new = RotXY(x,y, Angle ,NewOrigin)

    # Polynomial fit
    P_new = np.polyfit(x_new, y_new, N)
    P_new2 = [Val if 1e-15 < abs(Val) else 0 for Val in P_new]
#    P_new3 = [Val if 1e-15 < abs(Val) else 0 for Val in P_new]
    return P_new2

#================================================================
#  RotPoint
#================================================================
def RotPoint(XY, Theta = 0, CentreOfRotation = np.array([0,0])):
    '''
    The same as RotXY, but with input (x1, y1) intead of
    [x1...xn] , [y1 ... yn]
    '''
    (x,y) =  RotXY(XY[0], XY[1], Theta = Theta, CentreOfRotation = CentreOfRotation)
    return np.array([x, y])

#================================================================
#  RotVersor
#================================================================
def RotVersor(V, Angle, Deg = False):
    '''
    Rotate the versor V = (Vx, Vy)

    '''
    if Deg== True:
        Angle = Angle * np.pi/180
        
    if (Angle == 0) or np.linalg.norm(V)==0:
        return V
    else:
        U = RotXY(V[0], V[1], Angle)
        return np.array([U[0][0], U[1][0] ])

#================================
#  Normalize
#================================
'''
Normalize an array
'''
def Normalize(v):
    v = np.array(v)
    return v/norm(v)
#================================
#  UnitVectorNormal
#================================
def UnitVectorNormal(v, Sign = +1):
    '''
    Performs a rotation of Sign * pi/2 of v.

    >>> return  rm.RotXY(v, Sign * np.pi/2)
    '''
    nx,ny = RotXY(v[0], v[1], Sign * np.pi/2)

    nx = 0 if abs(nx) < 1e-16 else nx[0]
    ny = 0 if abs(ny) < 1e-16 else ny[0]
    return np.array([nx,ny])
#================================
#  UnitVectorReflect
#================================
def UnitVectorReflect(v, n):
    '''
    Parameters
    -------------
    v : vector-like (vx,vy)
        Unit vector of the incident beam
    n : vector-like
        normal of the mirror surface
 

    v and n should be unit vectors. If they are not, they are automatically
    normalized.
    '''
    v = np.array(v) # vector to be reflected
    n = np.array(n) # versor normal to the surface
    v = v / norm(v)
    n = n / norm(n)

    t = np.dot(n,v); # scalar product
#        print(v)
#        print(n)
#        print(norm(v))
#        print(norm(n))
#        print(t)
    if t >= 0:
      u = v;
    else:
      u = v - 2 * n * t;
    return u

#================================
#  FitSphericalWave
#================================
def FitSphericalWave1d(Phi,s, Lambda):
    '''
    
    Performs the quadratic fit over the phase Phi and returns the curvature 
    Radii.
    
    Interpolation kernel copied from:
    From: http://scipy-cookbook.readthedocs.io/items/Least_Squares_Circle.html
    
    Parameters
    -----
    Phi : array like
        phase of the spherical wave
        
    s : array
        coordinate points.

    '''
    x = s 
    y = Phi * Lambda/2/np.pi # convertion from phase to heigth profile
     
    from scipy import optimize
    method_2 = "leastsq"
    # coordinates of the barycenter
    x_m = np.mean(x)
    y_m = np.mean(y)

    def calc_R(xc, yc):
        """ calculate the distance of each 2D points from the center (xc, yc) """
        return np.sqrt((x-xc)**2 + (y-yc)**2)
    
    def f_2(c):
        """ calculate the algebraic distance between the data points and the mean circle centered at c=(xc, yc) """
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    
    center_estimate = x_m, y_m
    center_2, ier = optimize.leastsq(f_2, center_estimate)
    
    xc_2, yc_2 = center_2
    Ri_2       = calc_R(*center_2)
    R_2        = Ri_2.mean()                  #_2 is because this is method#2 from the url I copied from.
    residu_2   = sum((Ri_2 - R_2)**2)
#    residu2_2  = sum((Ri_2**2-R_2**2)**2)
#    ncalls_2   = f_2.ncalls        
    
    
    return R_2

#================================
#  FitGaussian1d
#================================
def FitGaussian1d(y, x = None, PlotFigure=None):
    '''
    Easy way for fitting gaussian curve.
    
    It's quick and dirty, but it should work
    
    Returns
    ----
    [a, x0, sigma] : Amplitude, mean value, standard deviation 
    '''
    from scipy.optimize import curve_fit
    from scipy import asarray as ar,exp
    # HELPER FUNCTION
    def gaus(x,a,x0,sigma):
        return a*exp(-(x-x0)**2/(2*sigma**2))
    #------------------------------------------
    n = len(y)
    if x is None:
        x = np.arange(0, n)
    
    mean = sum(x*y)/n    
    sigma = np.sqrt(sum(y*(x-mean)**2)/n)
    amplitude = max(y)
    
    popt,pcov = curve_fit(gaus,x,y,p0=[amplitude,mean,sigma])
    
    popt[2] = abs(popt[2])
    
    # Plot figure if required
    if PlotFigure!= None:
        if PlotFigure> 1:
            plt.figure(PlotFigure)
            plt.plot(x,y)
            xfit = np.linspace(min(x), max(x), 100)
            yfit = gaus(xfit,popt[0], popt[1], popt[2])
            plt.plot(xfit,yfit,'o')
        
    return popt
    #================================
    #  Coerce
   #================================
def Coerce(x, Min, Max):
    return Min if x < Min else Max if x > Max else x
    
#================================
#  L2XY
   #================================
def L_2_XY(f, L, XStart,  Sign = +1, Tolerance = 1e-3, GuessStep = 1e-4, iMax = 1e5):
    '''
    For a given function y = f(x) and a starting point XStart,
    it finds the point XEnd such that the length of the curve
    f(x) from XStart to XEnd equals L
     
    We have originally written it for accurately finding the XStart, XEnd points of 
    an elliptic mirror.
    
    '''

    # Single path integral
    XStep = GuessStep * Sign
    XOld = XStart
    YOld = f(XStart)
    i=0
    LSum = 0
    while True:
        XNew = XOld + XStep
        YNew = f(XNew)[0]
        DeltaL = norm(np.array([XNew, YNew])- np.array([XOld, YOld])) 
        LSum = LSum + DeltaL 
        
        i = i+1
        
        if LSum >= L: # raggiunta la lunghezza
            if (LSum - L) > Tolerance: 
                # ripeti con step più piccolo
                pass #???
            else:
                return (XNew, YNew)
        else: # da rifare
            if i >= iMax:
                return None
            pass
    pass
        
    





#================================
#  MinHew
#================================
def MinHew(Hew, Threshold = 1e-15):
    '''
    Rough way of finding the minimum value of the Hew, without fitting.
    Finds the minimum. If different values are repeated, then the central element
    is chosen.
    
    Equality is to the best of Threshold
    
    '''
    iMin = np.argmin(Hew) # position of the minimum
    Delta = Hew - Hew[iMin]  # differences.
    Bools = Delta < Threshold
    # find the first True value
    First = next(i for i in range(len(Bools)) if Bools[i]  == True)
    Last = next(i for i in np.arange(len(Bools)-1,-1,-1) if Bools[i]  == True)
    
    #Performs a quadratic fit
        
    return int(np.floor( (First + Last)/2))



#================================
#  FitParabola
#================================
def FitParabola(x, y ):
    """
    Helper function that performs a parabolic fit on x,y and returns
    the parabola parameter in a common fashion
    
    Parameters
    -----
    x : array
        x
    y : array
        y
        
    Return
    ------
    Struct containing:
        a,b,c,Focus, Vertex
    
    """
    
    class Output:
        a = None
        b = None
        c = None
        Focus = None
        Vertex = None
    
    #Performs a quadratic fit
    p = np.polyfit(x,y,2)
    a = p[0]
    b = p[1]
    c = p[2]
    Delta = b**2  -4*a*c
    
    xv = -b/2/a           # Vertex
    yv = -Delta/4/a
    
    xf = -b/2/a           # focus
    yf = (1-Delta)/a/a ; 
    
    Output.a = a 
    Output.b = b
    Output.c = c
    Output.Focus = np.array([xf,yf])
    Output.Vertex= np.array([xv,yv])
    
    return Output
#================================
#  FindWaist
#================================
def FindWaist(W, Z = None, Threshold = 1e-15):
    '''
    Find the waist of the beam of transverse size W and longitudinal axis Z.
    
    Parameters
    -----
    W : array like
        Beam size, typically the HEW (or the sigma, or whatever). It must have 
        a local minimum
        
    Z : array like
        longitudinal coordinate
    
    Return
    -----
    NumericWaist : (position, size) 
        coordinates of the minimum of (W,Z) found as the minimum value of W 
        
    FittedWaist : (position, size)
        coordinates of the minimum of (W,Z) computed with a parabolic fit
    
    '''
    iMin = np.argmin(W)            # position of the minimum
    Delta = W - W[iMin]        # differences.
    Bools = Delta < Threshold
    
    # find the first True value
    First = next(i for i in range(len(Bools)) if Bools[i]  == True)
    Last = next(i for i in np.arange(len(Bools)-1,-1,-1) if Bools[i]  == True)
    
    MinIndex = int(np.floor( (First + Last)/2))    # position of the central minimum
    MinValue = W[MinIndex]
    MinZ = Z[MinIndex] if Z != None else MinValue
    
    Fit = FitParabola(Z,W)
    
    MinFitValue = Fit.Vertex[1]
    MinFitZ = Fit.Vertex[0]
    
    NumericWaist = (MinZ, MinValue)
    FittedWaist = (MinFitZ, MinFitValue)
    
    return NumericWaist, FittedWaist
    
    
    


    
    
    
    
        
    
    return 



#================================
#  PhaseUnwrap
#================================        
#def PhaseUnwrap(Ph):
#    N = len(Ph)
#    PhNew = 0 * Ph
#    
#    Ph = Ph - Ph[0]
#    for (i,phi) in enumerate(Ph):
#        if i=0:
#            continue
#        else:
#            pass
#        
#    
#    
#================================
#  SlitCreate
#================================    
def SlitCreate(SizeN, NSlits = 1, SlitKernel = [1]):
    '''
    This function was first created for simulating the diffraction from 2 or more slits.
    
    Parameters
    ----
    SizeN : the total size (in samples) of the mask,
    
    NSlits : the total number of slits (1,2, etc...)
    
    SlitKernel : the transmission function of each slit. Can be complex.
    
    Example: create an array with 20 elements, 2 slits, each one 3 pixel large.
    SlitCreate(20,2, [1,1,1]).
    

    '''    
    
    fList =np.floor(np.linspace(0,SizeN, NSlits+2)[1:-1])
    iList = [int(f) for f in fList]
    A = np.zeros([SizeN])  # the total domain of the signal
    A[iList] = 1 # the 'lattice' position of the signal (e.g. locations of the slits)
    K = np.array(SlitKernel) # the kernel function (e.g. slit width)
    S = np.convolve(A,K)
    return S

class geom:

    #================================
    #  PointsToLine
    #================================
    @staticmethod
    def PointsToLine(x0,y0,x1,y1):
        '''
        Returns
        --------
        m, q
        Line equation y=m*x+q
        '''
        x = [x0,x1]
        y = [y0,y1]

        p = np.polyfit(x,y,1)
        q = p[0]
        m = p[1]
        return m,q


#    def PolarToLine(x0, y0, angle):
#        m = np.atan(angle)
#        return m,q

    #================================
    #  StepAlongLine
    #================================
    @staticmethod
    def StepAlongLine(step, m,q, x0 = 0 , sign = +1):
        '''
        Returns
        -----------------
        x_end, y_end
        '''
        y0 = m*x0 + q
        a = 1+m**2
        b = -2*x0 - 2*y0*m + 2*m*q
        c = -step**2 + x0**2 + y0**2 +q**2 - 2*y0*q

        x = (-b + sign* np.sqrt(b**2 - 4*a*c))/2/a
        y = m*x+q

        return x,y

     #================================
    #  StepAlongDirection
    #================================
    @staticmethod
    def StepAlongDirection(x0 , y0, step, angle,  sign = +1):
        '''
        Parameters
        ----------
        step : length of the movement
        angle : inclination angle (radians)
        x0, y:origin : points
        '''
        # horizontal displacement
        if (angle == 0) or (angle == np.pi) :
            if angle == np.pi:
                sign = sign * -1 ;
            x = x0 + step*sign
            y = y0

        # vertical displacement
        elif    (angle == np.pi/2) or (angle == - np.pi/2) :
            if angle == - np.pi/2:
                sign = sign* -1
            x = x0
            y = y0 + step*sign

        # generic displacement
        else:
            m = np.tan(angle)
            q =   -m* x0 + y0
            angle_mod = np.mod(angle, 2 * np.pi)
            if (0 <= angle_mod <= np.pi /2) or (3/2*np.pi < angle_mod <= 2 * np.pi):
                sign = 1*sign
#                print('sign kept')
            else:
                sign = -1*sign
#                print('inverto segno')

            x,y = geom.StepAlongLine(step, m,q,x0, sign)
        return x,y
    #================================
    #  IntersectLine
    #================================
    @staticmethod
    def IntersectLine(Poly1,Poly2):
        '''
        Intersects two lines.
        -------------
        Poly1 and Poly2  are in the form [m, q]
        where the line equation is :math:`y = m x + q`.

        Returns
        ------------
        x, y : scalars
            Point of intersection.
        '''
        x0 = (Poly2[1] - Poly1[1])/(Poly1[0] - Poly2[0])
        y0 = Poly1[0] * x0 + Poly1[1]

        return (x0,y0)



    #================================
    #  UnitVectorNormal
    #================================
    @staticmethod
    def UnitVectorNormal(v, Sign = +1):
        '''
        Performs a rotation of Sign * pi/2 of v.

        >>> return  rm.RotXY(v, Sign * np.pi/2)
        '''
        nx,ny = RotXY(v[0], v[1], Sign * np.pi/2)

        nx = 0 if abs(nx) < 1e-16 else nx
        ny = 0 if abs(ny) < 1e-16 else ny
#        print 10 * '===' 
#        print(nx)
#        print (ny)
        return np.array([nx,ny])

    #================================
    #  UnitVectorReflect
    #================================
    @staticmethod
    def UnitVectorReflect(v, n):
        '''
        Parameters
        -------------
        v : vector-like
            Unit vector of the incident beam
        n : vector-like
            normal of the mirror surface


        v and n should be unit vectors. If they are not, they are automatically
        normalized.
        '''
        v = np.array(v)
        n = np.array(n)
        v = v / norm(v)
        n = n / norm(n)

        t = np.dot(n,v);
#        print(v)
#        print(n)
#        print(norm(v))
#        print(norm(n))
#        print(t)
        if t >= 0:
          u = v;
        else:
          u = v - 2 * n * t;
        return u


#
#        Angle1 = np.arctan(m1)
#        # incident ray
#        Ray0 = geom.Ray(Angle0, [x0,y0])

#        # intersection
#        (x2, y2) = geom.IntersectLine([Ray0.m, Ray0.q], [m1,q1])
##        q3 = x2 * (m1**2 + 1)\m1 + q1
#
#        Angle2 = 0*np.pi - Angle0 + Angle1
#        return x2, y2, Angle2

    #================================
    #  DrawCircle
    #================================
    @staticmethod
    def DrawCircle(R,x0=0,y0=0,N=100):
        an = np.linspace(0, 2*np.pi, 100)
        return R*np.cos(an), R*np.sin(an)

    #================================
    #  DrawSegment
    #================================
    @staticmethod
    def DrawSegment(x0,y0,x1,y1, N=2):
        '''
        Returns
        --------------
        x,y : 1d arrays containing the points of the segment.
        '''
        x = [x0,x1]
        y = [y0,y1]
        p = np.polyfit(x,y,1)
        x = np.linspace(x0,x1,N)
        y = np.polyval(p,x)
        return x,y

    #================================
    #  DrawSegmentCentred
    #================================
    @staticmethod
    def DrawSegmentCentred(L, x0, y0, Angle, N=2):
        xEnd,yEnd =     geom.StepAlongDirection(L/2, Angle, x0,y0, sign=+1)
        xStart,yStart = geom.StepAlongDirection(L/2, Angle, x0,y0, sign=-1)
        x,y = geom.DrawSegment(xStart,yStart,xEnd,yEnd,N)
        return x,y




    


    
#    #==============================================================================
#    #     CLASS: Ray_old
#    #==============================================================================
#    class Ray_old(Line):
#        ''' Essentially, a Ray object contains the coeffcient of a line + info about
#            the origin of the Ray + info about the end of the Ray, in the case of a
#            focussed beam.
#            Angles are expressed in radians
#            Theta
#            q, XYStart,
#            Lenght
#            XYEnd
#        '''
#        #======================
#        #     __init__
#        #======================
#        def __init__(self, Angle = None, XYOrigin = np.array([0,0])):
#
#            if CheckArg([Angle,XYOrigin]):
#                self.Angle = Angle
#                self.XYOrigin = XYOrigin
#                self._UpdateVersor()
#            else:
#                pass
#
#
#        #======================
#        #     Angle
#        #======================
#        @property
#        def Angle(self):
#            return self._Angle
#        @Angle.setter
#        def Angle(self, Val):
#            if Val != None:
#                self._Angle= Val
#                self._m = np.tan(self._Angle)
#
#        #======================
#        #     m
#        #======================
#        @property
#        def m(self):
#            return self._m
#        @m.setter
#        def  m(self, val):
#            if val!=None:
#                self._m = val
#                self._Angle= np.arctan(val)
#
#        #======================
#        #     q
#        #======================
#        @property
#        def q(self):
#            return self._q
#        @q.setter
#        def q(self, val):
#            self._q = val
#
#
#        #======================
#        #     XYOrigin
#        #======================
#        @property
#        def XYOrigin(self):
#            return self._XYOrigin
#        @XYOrigin.setter
#        def    XYOrigin(self,XY):
#            if XYis None : exit
#            self._XYOrigin = XY
#            self._q = XY[1] - self.m * XY[0]
#
#        #======================
#        #     Draw
#        #======================
#        def Draw(self,L, N=100):
#            """
#            Returns the x,y arrays of points which can be used to plot the ray.
#
#            Paramteres
#            -----------------
#            L : scalar
#                Length of the ray to plot
#            N : int
#                Number of samples
#
#            Returns
#            -----------------
#            x : 1darray
#                List of x points
#            y : 1darray
#                List of y points
#
#            """
#
#            XYList = np.array([geom.StepAlongDirection(self.XYOrigin[0],
#                                    self.XYOrigin[1], L/i, self.Angle) for i in range(1,N+1)])
#            return XYList[:,0], XYList[:,1]
#
#    #================================
#    #  RayReflectFromLine
#    #================================
#    @staticmethod
#    def RayReflectFromLine(Ray_in = Ray(), Line_mirror = Line()):
#        '''
#        RayReflectFromLine(Ray_in = Ray(), Line_mirror = Line())
#
#        Calculates the reflection of an (oriented) ray onto a line.
#        Subscripts 0 refer to incident ray, 1 to reflecting line, 2 to reflected ray.
#
#        Parameters
#        ----------
#        Ray_in : Ray object
#            ray defining the incident bea,
#        Line:mirrror : Line object
#            Line defining the reflectiong "surface"
#
#        Returns
#        -------
#        Ray_out : Ray object
#            Ray defining the reflected beam
#        XYIntersection : 1x2 array
#            Contains the x,y coordinates of the intersection betweeen the line containing the
#            ray and the mirror line. If the ray is oriented such that the optical intersection
#            really occurs, the Intersect output is True. Else, it it is set to false
#        Intersects : bool
#            Says whether the ray impinges on the mirror line or not.
#        More
#        -------
#        This is the basis for the reflection operations.
#        It does not include
#        - check over the physical extension of the mirror
#        - reflection from "mirror" more other than rect.
#        '''
#
#        v_in = Ray_in.v
#        v_mir = Line_mirror.v
#
#        n_mir = rm.RotVersor(v_mir, np.pi/2)
#
#        scalar_product = np.dot(v_in, n_mir)
#        v_out = v_in - 2 * n_mir * scalar_product
#        Angle_out = np.arctan2(v_out[1], v_out[0])
#        # intersection
#        (out_x0, out_y0) = geom.IntersectLine([Ray_in.m, Ray_in.q],
#                                             [Line_mirror.m,Line_mirror.q])
#
#
#        Ray_out = geom.Ray(Angle = Angle_out, XYOrigin = [out_x0, out_y0])
#
#        #TODO:SERIOUS
#        Intersect = None
#        if ((out_x0 < np.sign(v_in[0])*Ray_in.XYOrigin[0]) or
#            (out_y0 < np.sign(v_in[1])*Ray_in.XYOrigin[1])):
#            Intersect = False
#        else:
#            Intersect = True
#        return Ray_out, out_x0, out_y0, Intersect
##----------------------------------------------------------


#==============================================================================
#     CLASS: Line
#==============================================================================
class Line(object):
    '''
    Attributes
    -------------------------
    m : scalar
        slope
    q : scalar
        intercept
    Angle : scalar (rad)
        rotation angle, defined as Angle = arctan(m)
    Versor : 1x2 array
        Versor defining the direction of the line
    VersorN : 1x2 array
        Versor defining the normal diretion to the line

    '''
    def __init__(self, m= None, q = None, Angle=None):

        if CheckArg([m,q]):
            self.q = q
            self.m = m
            self._UpdateVersor()
        elif CheckArg([Angle,q]):
            self.q = q
            self.Angle= Angle
            self._UpdateVersor()
        else:
            pass
        #calculates the versor according to the rotation angle used

    #======================
    #     Angle
    #======================
    @property
    def Angle(self):
        return self._Angle
    @Angle.setter
    def Angle(self, Val):
        self.Angle = Val
        self._m = np.tan(Val)

    #======================
    #     m
    #======================
    @property
    def m(self):
        return self._m
    @m.setter
    def  m(self, val):
        self._m = val
        self._Angle = np.arctan(val)

    #======================
    #     q
    #======================
    @property
    def q(self):
        return self._q
    @q.setter
    def q(self, val):
        self._q = val

    #======================
    #     v
    #======================
    @property
    def v(self):
        '''
        Versor defining the direction along the Line
        '''
        return self._Versor
    #======================
    #     vNorm
    #======================
    @property
    def vNorm(self):
        '''
        Versor defining the direction normal to the line
        '''
        return self._VersorN

    #======================
    #     _UpdateVersor
    #======================
    def _UpdateVersor(self):
        '''

        '''
        self._Versor = RotVersor([1,0], self.Angle)
        self._VersorN = RotXY(self._Versor, np.pi/2)

    #======================
    #     Draw
    #======================
    def Draw(self,xStart, xEnd, N=100):
        """
        Returns the x,y arrays of points which can be used to plot the Line.

        Paramteres
        -----------------
        xStart : scalar
            Start x Value
        xEnd : scalar
            End x value

        N : int
            Number of samples

        Returns
        -----------------
        x : 1darray
            List of x points
        y : 1darray
            List of y points

        """
        x = np.linspace(xStart, xEnd, N)
        y = self.m * x + self.q
        return x,y

#==============================================================================
#     CLASS: Segment
#==============================================================================
class Segment(Line):

    #======================
    # FUN: __init__
    #======================
    def __init__(self, x0,y0,x1,y1):
        m,q = geom.PointsToLine(x0,y0,x1,y1)
        Line.__init__(self,m=m, q=q)
        self._XYStart = np.array([x0,y0])
        self._XYEnd = np.array([x1,y1])

    #======================
    # PROP:     XYStart
    #======================
    @property
    def XYStart(self):
        return self._XYStart
    @XYStart.setter
    def XYStart(self,val):
        self._XYStart = val
        # "joins" the new XYStart with the stored XYEnd, and updates the other
        # variables
        m,q = geom.PointsToLine(self._XYStart[0], self._XYStart[1],                                                 self._XYEnd[0], self._XYEnd[1])

    #======================
    # PROP:     XYEnd
    #======================
    @property
    def XYEnd(self):
        return self._XYEnd
    @XYEnd.setter
    def XYEnd(self,val):
        self._XYEnd = val
        # "joins" the stored XYStart with the NEW XYEnd, and updates the other
        # variables
        m,q = geom.PointsToLine(self._XYStart[0], self._XYStart[1],                                                 self._XYEnd[0], self._XYEnd[1])
        self.q = q
        self.m = m

    #======================
    # METHOD:     __Draw
    #======================
    def __Draw(self,N=2):
#            return geom.DrawSegment(XYStart[0],XYStart[0], XYEnd[0],XYEnd[1])
        pass
#==============================================================================
#     CLASS: Vector
#==============================================================================
class Vector(object):
    '''
    The unit vector is automatically normalized to unit norm, basing on the inputs 
    parameters.
    
    Set of Parameters 1
    -------------------
    x : x component
        
    y : y component
    
    Set of Parameters 2
    -------------------
    Angle : Angle (radians)

    Set of Parameters 3
    -------------------
    V : a Vector object

    Set of Parameters 4
    -------------------
    v : a v = [vx,vy] type
        
    '''

    #================================
    # FUN: __init__
    #================================
    def __init__(self, 
                vx = None, vy = None, 
                v = None, # [vx,vy] array
                V = None, # a vector object
                Angle = None, 
                XYOrigin = [0,0], Length = 1, IsUnitVector = False):
    
        self._IsUnitVector = IsUnitVector
        self.XYOrigin = XYOrigin         # Set it now, not later
        
        if CheckArg([vx,vy]):
            if (vx!=0 or vy !=0):
                self.v = [vx,vy] # _v is assigned
            else:
                self._ZeroLength() # _v is assigned
        elif CheckArg([v]):
            self.v = v          # _v is assigned
            
        elif CheckArg([Angle, Length]):    
            if Length >0:
                self.v = RotVersor(np.array([1,0]) *Length, Angle)  # _v is assigned
            else:
                self._ZeroLength()    # _v is assigned
        else: 
            print ('Vecor Init Wrong Argument set')

        
        
        if self.Length > 0:
            self._vNorm = UnitVectorNormal(self.v)
        else:
            self._vNorm = None

        if IsUnitVector == True:
            self.v = Normalize(self.v)


    #======================
    # FUN: __str__
    #======================
    def __str__(self):
        StrMsg = (' V = [%f,%f]\n Angle=%0.2f  deg\n XYOrigin=[%f,%f]' %     
                (self.v[0], self.v[1], (self.Angle * 180/np.pi), 
                self.XYOrigin[0], self.XYOrigin[1]))
        return StrMsg
    #======================
    # FUN: _ZeroLength
    #======================
    def _ZeroLength(self):
        '''
        The Vector is recognized to be degenerate; hence I set all the values accordingly.
        '''
        self._v = np.array([0,0])
        self.vAngle = 0
        self._vNorm = None
        self.XYEnd = self.XYOrigin
        
    #======================
    # PROP: v
    #======================
    '''
    v: is the numerical array containing the components
    '''
    @property
    def v(self):
        return self._v
    @v.setter
    def v(self, value):
        vx = value[0]
        vy = value[1]
        # Normalize to unity?
        L2 = 1 if self._IsUnitVector == False else      vx**2 + vy**2
        self._v = np.array([vx/L2, vy/L2])
        self._vAngle = np.arctan2(self._v[1],self._v[0])
        
    #======================
    # PROP: vAngle
    #======================
    @property
    def vAngle(self):
        return self._vAngle
    @vAngle.setter
    def vAngle(self, value):
        self._v = RotVersor(np.array([1,0]) *self.Length, value)  # don't use self.Lenght (it bases on ._v)
        self._vAngle = value    
        
    #======================
    # PROP: Angle
    #======================
    @property
    def Angle(self):
        return self._vAngle
    @vAngle.setter
    def Angle(self, value):
        self._v = RotVersor([1,0], value)
        self._vAngle = value
    
    #======================
    # PROP: PolyCoeff
    #======================    
    @property
    def PolyCoeff(self):
        '''
        Returns the coefficients of the 1st degree polynomial p =[m,q] (line equation)
        '''
        XYEnd = self.XYOrigin + self.v
        x = np.array([self.XYOrigin[0], XYEnd[0]])
        y = np.array([self.XYOrigin[1], XYEnd[1]])
        return np.polyfit(x,y,1)
    #======================
    # PROP: vNorm
    #======================
    @property
    def vNorm(self):
        return self._vNorm
    #======================
    # PROP: XYOrigin
    #======================
    @property
    def XYOrigin(self):
        return self._XYOrigin
    @XYOrigin.setter
    def XYOrigin(self, value):
        self._XYOrigin = np.array(value)
        
    #======================
    # PROP: XYEnd
    #======================
    @property
    def XYEnd(self):
        return self._XYOrigin + self._v
    @XYEnd.setter
    def XYEnd(self, value):
        self._Length = norm(value - self._XYOrigin)

    #======================
    # PROP: Length
    #======================
    @property
    def Length(self):
        try:
             return np.linalg.norm(self._v)
        except:
            return 0
    #======================
    # METHOD: Rotate
    #======================
    def Rotate(self, Angle):
        '''
        Rotates the present Unit vector object by angle
        It is equivalent to the code
        
        >>> UnitVector.vAngle +=  Angle
        '''
        self.vAngle = self.vAngle + Angle


    #======================
    # PROP: VersorNor
    #======================
    @property
    def VersorNorm(self):
        ''' Returns a UnitVector object which is normal to this objkect'''
        import copy
        _Versor = copy.deepcopy(self)
        _Versor.Rotate(+np.pi/2)
        return _Versor
        
    #======================
    # METHOD: GetNormal
    #======================
    def GetNormal(self):
        '''
        Returns another vector which is normal to the present one.
        It is equivalent to rotate of +pi/2
        
        >>> UnitVector.vAngle +=  Angle
        '''
        import copy
        _v = copy.deepcopy(self)
        _v.Rotate(+np.pi/2)
        return _v
    #======================
    # METHOD: Paint
    #======================        
    def Paint(self, FigHandle = None,  Length = 0.3, ArrowWidth = 0.3,
           Color = 'k', Shift = False):
        plt.figure(FigHandle)
        ax = plt.axes()
        ArrowLength = 1.5 * ArrowWidth
        XYEnd = self.XYOrigin +  self.v
        ShiftX = 0 if Shift == False else -np.cos(self.Angle)*(Length + ArrowLength)
        ShiftY = 0 if Shift == False else -np.sin(self.Angle)* (Length+ArrowLength)
        x0 = self.XYOrigin[0] + ShiftX 
        y0 = self.XYOrigin[1] + ShiftY
        x1 = XYEnd[0] + ShiftX
        y1 = XYEnd[1] + ShiftY
            
        ArrowWidth = 0.5 * Length if ArrowWidth is None else ArrowWidth
            
        ax.arrow(x0,y0,Length*(x1-x0), Length*(y1-y0), 
                   head_width = ArrowWidth, head_length= ArrowLength, 
                                   fc=Color, ec=Color)
#============================================================================
#     CLASS: UnitVector
#==============================================================================
class UnitVector(Vector):
    def __init__(self,  vx = None, vy = None, 
                v = None,
                V = None,
               Angle = None, XYOrigin = [0,0]):
        Vector.__init__(self,  vx = vx, vy= vy, v = v, V = V , 
                  Angle = Angle, XYOrigin = XYOrigin,
                  IsUnitVector = True)

#============================================================================
#     CLASS: Ray
#==============================================================================
class Ray(Vector):
    #======================
    # FUN: __init__
    #======================
    def __init__(self, x0 = None,y0 = None ,x1 = None ,y1 = None,
                    XYOrigin = None, Angle = None, FocalLength = float('inf'), Length = 1,
                    vx = None, vy = None    ):
        '''
        Parameters Set
        ---------------------------------

        '''
        
        # Parameter Set 1
        if CheckArg([x0,y0,x1,y1]):
            # Store input
            Vector.__init__(self, vx = x1-x0, vy = y1-y0, IsUnitVector = True )

        # Parameter Set 2
        elif CheckArg([XYOrigin, Angle]):
            # Store input
            Vector.__init__(self, Angle=Angle, XYOrigin = XYOrigin, IsUnitVector = True  )
            self.XYOrigin = XYOrigin
        elif CheckArg([vx, vy]):
            Vector.__init__(self, vx = vx, vy = vy, IsUnitVector = True )
        elif all([Arg is None for Arg in [x0,y0,x1,y1,XYOrigin, Angle,FocalLength]]):
#                print('Ray.__init__ : empty Ray was initialized')
#                raise ValueError('A very specific bad thing happened')
            pass
        else:
            print('Ray.__init__ : a wrong combination of arguments was used')
            raise ValueError('A very specific bad thing happened')
             # Update further parameters (common procedure)
             
        self.XYOrigin = [0,0] if XYOrigin is None else XYOrigin
        self.FocalLength = FocalLength

    #======================
    # PROP: UnitVectorAtOrigin
    #======================        
    @property
    def UnitVectorAtOrigin(self):
        return UnitVector(v = self.v)


    #======================
    # PROP: Length
    #======================        
    @property
    def Length(self):
        return norm(self.v)

    #======================
    # PROP: Norm
    #======================        
    @property
    def Norm(self):
        return norm(self.v)
    #======================
    # PROP: XYEnd
    #======================        
    @property
    def Norm(self):
        return norm(self.v)

    #======================
    # PROP: FocalLength
    #======================        
    @property
    def FocalLength(self):
        return norm(self._FocalLength)
    @FocalLength.setter
    def FocalLength(self, value):
        value = float('inf') if value is None else value
        self._FocalLength = value
        self._XYFocus = self._XYOrigin + self._FocalLength * Normalize(self.v) if self._FocalLength < float('inf') else float('inf')
    
    #======================
    # METHOD: Paint
    #======================        
    def Paint(self, FigHandle = None, Length = None, ArrowWidth = 0.3, 
               Color = 'k', Shift = False):
        
        UV = UnitVector(v = self.v, XYOrigin = self.XYOrigin)
        _Length = Length = Length if Length != None else self.Length
        UV.Paint(FigHandle, _Length,  
                      ArrowWidth = ArrowWidth, Color = Color, Shift = Shift)
        
    #======================
    #     Draw
    #======================
    def Draw(self,L, N=100):
        """
        Returns the x,y arrays of points which can be used to plot the ray.

        Paramteres
        -----------------
        L : scalar
            Length of the ray to plot
        N : int
            Number of samples

        Returns
        -----------------
        x : 1darray
            List of x points
        y : 1darray
            List of y points

        """

        XYList = np.array([geom.StepAlongDirection(self.XYOrigin[0],
                                self.XYOrigin[1], L/i, self.Angle) for i in range(1,N+1)])
        return XYList[:,0], XYList[:,1]
        
#============================================================================
#     CLASS: Ray
#==============================================================================
class Ray_seminuovo(object):
    ''' Essentially, a Ray object contains the coeffcient of a line + info about
    Theta
    q, XYStart,
    Lenght
    XYEnd
    '''

    #======================
    # FUN: __init__
    #======================
    def __init__(self, x0 = None,y0 = None ,x1 = None ,y1 = None,
                    XYOrigin = None, Angle = None, FocalLength = None):
        '''
        Parameters Set
        ---------------------------------

        '''
        
        self.v = [1,0]
        self.vNorm =[0,1]

        # Parameter Set 1
        if CheckArg([x0,y0,x1,y1]):
            # Store input
            self.XYOrigin = np.array([x0,y0])
            self.XYEnd = np.array([x1,y1])

            # Update other parameters    (specific)
            dx = x1-x0
            dy = y1-y0
            L2 = dx**2 + dy**2
            self._v = np.array([dx/L2, dy/L2])
            self.m, self.q = geom.PointsToLine(0,0,self.v[0],self.v[1])
            self.Angle = np.arctan(self.m)

        # Parameter Set 2
        elif CheckArg([XYOrigin, Angle]):
            # Store input
            self.Angle = Angle
            self.XYOrigin = XYOrigin

            # Update other parameters (specific)
            self.v = RotVersor([1,0], Angle)
            self.m = self.v[1]/self.v[0]
            self.q = XYOrigin[1] - self.m * XYOrigin[0]
        elif all([Arg is None for Arg in [x0,y0,x1,y1,XYOrigin, Angle,FocalLength]]):
#                print('Ray.__init__ : empty Ray was initialized')
#                raise ValueError('A very specific bad thing happened')
            pass
        else:
            print('Ray.__init__ : a wrong combination of arguments was used')
            raise ValueError('A very specific bad thing happened')
             # Update further parameters (common procedure)
        self._vNorm = UnitVectorNormal(self.v)

    #======================
    # PROP: Length
    #======================        
    @property
    def Length(self):
        return norm(self.v)

    #======================
    # PROP: Norm
    #======================        
    @property
    def Norm(self):
        return norm(self.v)
    #======================
    # PROP: XYEnd
    #======================        
    @property
    def Norm(self):
        return norm(self.v)

        
    #======================
    # METHOD: Paint
    #======================        
    def Paint(self, FigHandle = None, ArrowWidth = 0.3):
        V = UnitVector(Angle = self.Angle, XYOrigin = self.XYOrigin)
        V.Paint(FigHandle, Length = self.Length,  ArrowWidth = ArrowWidth)
        
    #======================
    #     Draw
    #======================
    def Draw(self,L, N=100):
        """
        Returns the x,y arrays of points which can be used to plot the ray.

        Paramteres
        -----------------
        L : scalar
            Length of the ray to plot
        N : int
            Number of samples

        Returns
        -----------------
        x : 1darray
            List of x points
        y : 1darray
            List of y points

        """

        XYList = np.array([geom.StepAlongDirection(self.XYOrigin[0],
                                self.XYOrigin[1], L/i, self.Angle) for i in range(1,N+1)])
        return XYList[:,0], XYList[:,1]

#    def DrawCircle(x,y,R, x0, y0)
#%%
#import matplotlib.pyplot as plt
#plt.close('all')
#plt.axis('equal')
#xc,yc = geom.DrawCircle(1)
#plt.plot(xc,yc)
#plt.xlim(-5,5)
#plt.ylim(-5,5)
#

